"""
Open Handhelds Browser

this program is meant to be used on the Caanoo Handheld, it browses 
the Open Handhelds archive for the Caanoo (www.openhandhelds.org) - at 
least in its current form - using simple heuristics and can install apps 
directly from the caanoo but only if:
1) the archive file is in a zip,rar or tar.gz, 7z (limited) file 
2) the archive contains precisely 1 .ini file with a matching directory

any additional libraries contained in the archive will not be extracted.

Note that this program will write on the SD card, as such it can potentially
ruin things badly (it has some basic sanity checks, though), in other words:
USE AT YOUR OWN RISK!

it needs the python+pygame+albow package

start up the network (wifi or usb) before starting the browser

v2: added rar and tar.gz support and added fixes to remedy somewhat crashes that occur 
when doing multiple things simultaneously

v3: improved responsiveness; ignores ini files in deeper levels; recent and popular menu;
bit more polish;

v4: (limited) 7z support, overwrite option, screenshot preview, fix of issue with games that 
do no show up in menu, cancellable install 

Freeware

included (check for respective licenses in LICENSE/):
 unrar: www.rarlab.com
 python rarfile: http://pypi.python.org/pypi/rarfile/2.3
 pylzma: http://www.joachim-bauch.de/projects/pylzma/
 bzip2: bzip.org

"""

import os, sys
import shutil
import urllib
from cStringIO import StringIO
import re
import threading
lock=threading.Lock()
#import zipfile
#import rarfile


from os.path import dirname as d
sys.path.insert(1, d(d(os.path.abspath(sys.argv[0]))))

import pygame
pygame.init()

from albow.dialogs import alert
from albow.shell import Shell
from albow.screen import Screen
from albow.table_view import TableView, TableColumn
from albow.resource import get_font
from albow.controls import Image,Label, Button, RadioButton, CheckBox
from albow.layout import Row, Column

screen_size = (320, 240)
frame_time = 125 # ms

baseurl="http://dl.openhandhelds.org/cgi-bin/caanoo.cgi?"
home="0,0,0,0,1"

titlefont=get_font(16, "../../LiberationSans-Regular.ttf")
buttonfont=get_font(12, "../../LiberationSans-Regular.ttf")
textfont=get_font(10, "../../LiberationSans-Regular.ttf")
messfont=get_font(9, "../../LiberationSans-Regular.ttf")

#targetdirs=['/tmp/game','/tmp/app']
targetdirs=['/mnt/sd/apps','/mnt/sd/game']

def urlretrieve(url,outfile,reporthook=None,cancelhook=None, chunk_size=131072):
  u=urllib.urlopen(url)
  total_size = u.info().getheader('Content-Length').strip()
  total_size = int(total_size)
  f=open(outfile,'wb')
  bytes_so_far = 0
  while 1:
    if cancelhook:
      if cancelhook():
        break
    chunk = u.read(chunk_size)
    bytes_so_far += len(chunk)

    if not chunk:
      break

    if reporthook:
      reporthook(bytes_so_far/chunk_size, chunk_size, total_size)
    f.write(chunk)
  f.close()  

class Banner(object):
  texts=[]
  text=''
  def __init__(self, text):
    self.texts.append(text)
    self.message(self.texts[-1])
  def message(self,text):
    f = messfont
    w=f.size(text)[0]
    if(w <= 290):
      s=int((290-w)/f.size(' ')[0])*' '
      globaltext=text+s
    if(w > 290):
      globaltext=text[:-int((w-290)/f.size('m')[0])]  
    self.texts[-1]=globaltext
    self.text=self.texts[-1]
  def downloadreport(self, block,blocksize,filesize):
    self.message("downloading %i%%"%(100*block*blocksize/filesize))

banner=Banner('visit www.openhandhelds.org and shop.gp2x.de')

class GlobalLabel(Label):
     font=messfont
     pos=0
     def draw(self,surf):
        global banner
        self.text=banner.text[self.pos:]+banner.text[:self.pos]
        self.pos=(self.pos+1) % len(banner.text)
        Label.draw(self,surf)

def info_from_download_url(url):
  with lock:
    try:
      a=urllib.urlopen(baseurl+url)
    except:
      banner.message('network trouble')
      raise Exception
    site_as_string=a.read()

  pat_download=re.compile("HREF='\?([^'>]*)'>Click Here")
  pat_filename=re.compile("Downloading <A HREF=[^><]*>([^<]*)<")
  pat_filename2=re.compile("Name:</B>...([^<]*)</FONT>",re.DOTALL)
  
  try:
    download=pat_download.findall(site_as_string)[0]
    filename=pat_filename.findall(site_as_string)[0]
  except:
    download=url.replace('0,1,0','0,1,2',1)
    filename=pat_filename2.findall(site_as_string)[0]

  return filename, download

def pull_image(url):
  with lock:
    try:
      a=urllib.urlopen(url)
    except:
      banner.message('network trouble')
      raise Exception
    s=StringIO(a.read())

  return pygame.transform.smoothscale(pygame.image.load(s).convert(),(160,120))

def pull_image_info(app_dict,url,image):
  if not app_dict.has_key('img'):
    app_dict['img']=pull_image(url)
  image.image=app_dict['img'] 
 
def fillappdictfromurl(app_dict,url):
  with lock:
    try:
      a=urllib.urlopen(url)
    except:
      banner.message('network trouble')
      raise Exception
    s=StringIO(a.read())
  lines=s.readlines()

  pat_title=re.compile('<TD VALIGN="TOP" WIDTH="40%"><FONT CLASS="big">&nbsp;<B>([^<]*)</B> </FONT>')
  desc_offset=10
  markup=re.compile('\<[^>]*\>')
  pat_break=re.compile('\<BR\>')

  prop=["Name","Date","Size","Version","Author","Licence"]
  offset=1

  for p in prop:
    pat=re.compile('\<B\>'+p+':')
    line=filter(lambda x: pat.search(x[1]),enumerate(lines))
    if len(line)!=1:
      banner.message('file info corrupt1')
      raise Exception
    app_dict[p]=markup.sub('',lines[line[0][0]+offset]).strip()
  
  pat_app=re.compile("HREF='\?([^'>]*)'>Download[^s:]")

  title=filter(lambda x:pat_title.search(x[1]),enumerate(lines))
  if len(title)!=1:
      banner.message('file info corrupt2')
      raise Exception
  
  app_dict['title']=pat_title.findall(title[0][1])[0]

  pat_img=re.compile('<TD WIDTH="50%" VALIGN="MIDDLE" ALIGN="CENTER">')
  img_offset=7
  imgsrc=re.compile(r"IMG SRC=\'([^']*)\'")

  img=filter(lambda x:pat_img.search(x[1]),enumerate(lines))
  if len(title)!=1:
      banner.message('info corrupt - if this persists check for updates')
      raise Exception

  try:
    app_dict['imgsrc']=imgsrc.findall(lines[img[0][0]+img_offset])[0]
  except:
    banner.message('info corrupt - if this persists check for updates')
    raise Exception
    
  download=filter(lambda x:pat_app.search(x[1]),enumerate(lines))
 
  if len(download)!=1:
      banner.message('file info corrupt3')
      raise Exception

  download=pat_app.findall(download[0][1])[0] 

  filename,downloadurl=info_from_download_url(download)
  
  app_dict['filename']=filename
  app_dict['downloadurl']=downloadurl

def filldictpopularrecent(rec_dict,pop_dict,url):
  with lock:
    try:
        a=urllib.urlopen(url)
    except:
      pop_dict['0No network']=('misc',cat_dict,'','connection problem; check your network. is the WIFI on?','')
      rec_dict['0No network']=('misc',cat_dict,'','connection problem; check your network. is the WIFI on?','')
      return
    s=StringIO(a.read())

  lines=s.readlines()
  
  markup=re.compile('\<[^>]*\>')
  recent_pat=re.compile('^(<TD><IMG SRC="http://dl.openhandhelds.org/caanoo/themes/carrick//images/arrow.gif" WIDTH="7" HEIGHT="7" BORDER="0"><FONT CLASS="small">)')
  pat_links=re.compile(r"HREF=['\"]\?([0-9,]+)['\"]")

  recent_offset=1

  recent=filter(lambda x: recent_pat.search(x[1]),enumerate(lines))

  def recentf(x):
    return pat_links.findall(x[1])[0],  \
         markup.sub('',x[1]).strip(), \
         markup.sub('',lines[x[0]+recent_offset]).strip()
  recent=map(recentf,recent)

  pop=filter(lambda x: x[2].find('/')==-1,recent)
  recent=filter(lambda x: x[2].find('/')>=0,recent)

  for i,c in enumerate(pop):
     pop_dict['%3.3i'%i+c[1]]=('app',dict(),c[0],c[1],'Download count: '+c[2])  
  for i,c in enumerate(recent):
     rec_dict['%3.3i'%i+c[1]]=('app',dict(),c[0],c[1],'Upload or update on '+c[2])  
  
  
def filldictfromurl(cat_dict,url,firstcall=True):
  with lock:
    try:
        a=urllib.urlopen(url)
    except:
      cat_dict['000No network']=('misc',cat_dict,'','connection problem; check your network. is the WIFI on?','')
      return
    s=StringIO(a.read())

  lines=s.readlines()

  pat_cat=re.compile('^(<TD WIDTH="190"><FONT CLASS="rspmid">)')
  pat_app=re.compile('^(<TD align="left"><FONT CLASS="mid">)')
  pat_page=re.compile('^(<TD VALIGN="MIDDLE" ALIGN="RIGHT" WIDTH="50%"><FONT CLASS="small"><B>Page 1,)')
  pat_links=re.compile(r"HREF=['\"]\?([0-9,]+)['\"]")
  pat_appname=re.compile(r'<A  HREF=[^>]*>([^<]*)</A>')
  catdes_offset=7
  appdes_offset=9
  markup=re.compile('\<[^>]*\>')

  cats=filter(lambda x: pat_cat.search(x[1]),enumerate(lines))
  apps=filter(lambda x: pat_app.search(x[1]),enumerate(lines))
  pages=filter(lambda x: pat_page.search(x[1]),enumerate(lines))

  if len(pages) > 2:
    raise Exception

  def catf(x):
    return pat_links.findall(x[1])[0],  \
           markup.sub('',x[1]).strip(), \
           markup.sub('',lines[x[0]+catdes_offset]).strip()
  cats=map(catf,cats)

  def appf(x):
    return pat_links.findall(x[1])[0],  \
           pat_appname.findall(x[1])[0].strip(), \
           markup.sub('',lines[x[0]+appdes_offset]).strip()
  apps=map(appf,apps)

  for c in cats:
     cat_dict['001'+c[1]]=('cat',dict([('000<- Up',('cat',cat_dict,'','move up',''))]),c[0],c[1],c[2])  
  for c in apps:
     cat_dict['002'+c[1]]=('app',dict(),c[0],c[1],c[2])  

  if len(pages)>=1 and firstcall:
    pageurl=pat_links.findall(pages[0][1])
    for u in pageurl:
      filldictfromurl(cat_dict,baseurl+u,firstcall=False)

def load_pop_recent(pop,rec,url,table_data):        
  table_data[:]=[' ..loading..']      
  if len(cur.keys())<=1 or cur.has_key('0No network'):
    if cur.has_key('0No network'):
      cur.pop('0No network')
    filldictpopularrecent(pop,rec,url)   
  table_data[:]=[]                  
  table_data.extend(sorted(cur.keys()))
    
        
def switchcategory(cur,url,table_data):
  table_data[:]=[' ..loading..']                  
  if len(cur.keys())==1 or cur.has_key('0No network'):
    if cur.has_key('0No network'):
      cur.pop('0No network')
    filldictfromurl(cur,url)   
  table_data[:]=[]                  
  table_data.extend(sorted(cur.keys()))


def appinfo(app_dict,url,val,title,image):
   global default_image
   val.text=' ..loading..'
   image.image=default_image
   if len(app_dict.keys()) ==0:
     fillappdictfromurl(app_dict,url)
   
   threading.Thread(target=pull_image_info, 
      args=((app_dict,app_dict['imgsrc'],image))).start()         
   val.text=''
   for p in ["Name","Date","Size","Version","Author","Licence"]:
     val.text+=p+':'+app_dict[p][:50]+'\n'
   title.text=app_dict['title'][:50]

cancel=False

def cancel_status():
  return cancel

def check_cancel(app):
  global cancel
  if cancel:
    banner.message('canceling install of '+app)
    cancel=False
    raise Exception

def install_app(app,targetdir,filename,url,overwritemode=False):
  global globaltext
  tmpdir='./tmp'+str(hash((filename,url)))

  os.makedirs(tmpdir)

  try:
    ext=filter(lambda x: filename.lower().endswith(x),['.zip','.rar','.tar.gz','.7z'])
    if not ext:
      ext=os.path.splitext(filename)[1].lower()
      banner.message('install of '+app+' aborted:'+ext+' not implemented')
      raise Exception
    ext=ext[0]

    banner.message('downloading '+app+'...'  )

    tmpfile=tmpdir+'/'+str(hash((filename,url)))+ext
    with lock:
      try:
        urlretrieve(baseurl+url,tmpfile,banner.downloadreport, cancel_status)
      except:
        banner.message('downloading ' + app + ' fails')   
    check_cancel(app)
    if ext=='.tar.gz':
      import tarfile
      try:
        archive=tarfile.open(tmpfile)
        archive.namelist=archive.getnames
      except:
        banner.message('install of '+app+' aborted: opening tar.gz failed')
        raise Exception
    elif ext=='.7z':
      import py7zlib
      try:
        archive=py7zlib.Archive7z(open(tmpfile,'rb'))
        archive.namelist=archive.getnames
      except:
        banner.message('install of '+app+' aborted: opening 7z failed')
        raise Exception
    elif ext=='.zip':
      import zipfile
      try:
        archive=zipfile.ZipFile(tmpfile)
      except:
        banner.message('install of '+app+' aborted: opening zip failed')
        raise Exception
    elif ext=='.rar':
      import rarfile
      rarfile.PATH_SEP='/'
      rarfile.UNRAR_TOOL = "./unrar"
      try:
        archive=rarfile.RarFile(tmpfile)
      except:
        banner.message('install of '+app+' aborted: opening rar failed')
        raise Exception
    else:
      banner.message('internal error: should not happen')
      raise Exception
    try:
      nl=archive.namelist()
    except:
      banner.message('no namelist from archive')
    ini=filter(lambda x: os.path.splitext(x)[1].lower()=='.ini' ,nl)

    check_cancel(app)
    if len(ini)==0:
      banner.message('install of '+app+' aborted: no ini file')
      raise Exception
    
    mindepth=min(map(lambda x: len(re.findall('/+',x)),ini))
    ini=filter( lambda x: len(re.findall('/+',x))==mindepth, ini)   

    if len(ini)>1:
      banner.message('install of '+app+' aborted: multiple ini files found')
      raise Exception
 
    inifile=ini[0]
    s=os.path.split(inifile)
    base=os.path.splitext(s[1])[0]
    filebase=os.path.splitext(inifile)[0]
    select=filter(lambda x: x.startswith(filebase+'/'),nl)
    if len(select)==0:
      banner.message('install of '+app+' aborted: directory no match ')
      raise Exception

    if not overwritemode:
      if os.path.exists(targetdir+'/'+base+'.ini'):
        banner.message('install of '+app+' aborted: ini file exists')
        raise Exception
      if os.path.exists(targetdir+'/'+base):
        banner.message('install of '+app+' aborted: target dir exists')
        raise Exception
      
    select.append(inifile)

    check_cancel(app)
    banner.message(' extracting archive...')
    if ext=='.7z':
      banner.message(' extracting 7z archive (may be slow)...')

    try:
      if ext=='.tar.gz':
        toextract=[archive.getmember(x) for x in select]
      else:
        toextract=select
      archive.extractall(tmpdir,toextract)      
    except:
      banner.message('install of '+app+' aborted: extract failed')
      raise Exception
    check_cancel(app)
    for x in select:
      try:  
        os.chmod(tmpdir+'/'+x,0755)
      except:
        pass
#      banner.message('install of '+app+' aborted: set permission fail')
#      raise Exception

    if overwritemode:
      if os.path.exists(targetdir+'/'+base+'.ini')!=os.path.exists(targetdir+'/'+base):
        banner.message('install of '+app+' aborted: existing ini without dir or vv')
        raise Exception  

      if os.path.exists(targetdir+'/'+base+'.ini'):
        import time
        for i in range(5):
          banner.message('removing '+targetdir+'/'+base+' in %i'%(5-i))
          time.sleep(1)
        check_cancel(app)

        os.remove(targetdir+'/'+base+'.ini')
        shutil.rmtree(targetdir+'/'+base)    

    check_cancel(app)
    try:
      shutil.move(tmpdir+'/'+inifile, targetdir)
      shutil.move(tmpdir+'/'+filebase,targetdir)
    except:
      banner.message('install of '+app+' fail')
      raise Exception

    banner.message(' install of '+app+' done!')

    if re.sub(base+'$','bgd-runtime', filebase) in nl:
      banner.message(' install done! (note: bgd-runtime must be present)')
         
  finally:
    shutil.rmtree(tmpdir)
    global busy_installing
    busy_installing=None

class installto(object):
  x=targetdirs[1]
  def set(self,x):
    self.x=x
  def get(self):
    return self.x

class overwrite(object):
  x=False
  def set(self,x):
    self.x=x
  def get(self):
    return self.x

class DangerCheckBox(CheckBox):
  def draw(self,surf):
    if self.ref.get():
      self.fg_color=pygame.Color('red')
    else:
      self.fg_color=pygame.Color('white')
    CheckBox.draw(self,surf)
      

default_image=pygame.Surface((160,120))

class AppInstallScreen(Screen):
	def __init__(self, shell):		
                Screen.__init__(self, shell)
		title = Label(appscreentitle, width=280, font = titlefont,align='c')
                self.title=title
                
                body = Label("install to:", font = buttonfont)
                choices=[body]
                self.i=installto()                
                for t in targetdirs:
                   choices.append( Row([RadioButton(ref=self.i,setting=t),Label(t, font = textfont)]) )
                val=Label(appinfotext,width=280,font=messfont,align='c')
                self.val=val
                
                self.image=Image(default_image)
                commit = Button("Commit", font=buttonfont,action = self.start_install ,enable=self.is_ready)
                back = Button("Back", font=buttonfont, action = shell.show_menu)
                glab=GlobalLabel('',width=290)
                self.commitorcancel=commit
                self.overwrite=overwrite()
                overwritecheck=Row([Label("Overwrite existing?", font = textfont), DangerCheckBox(ref=self.overwrite)])
                
                choices.append(overwritecheck)
                choices.append(Row([commit,back]))
                
                
                content=Column([title,self.val,Row([self.image,Column(choices,align='l')])], spacing=2,height=215)
                
                screen = Column([content,glab], spacing = 8)
		self.add_centered(screen)
                self.shell=shell
        def draw(self,surf):
          global busy_installing
          if busy_installing==app_to_install:
            self.commitorcancel.text='Cancel'
          else:
            self.commitorcancel.text='Commit'
          Screen.draw(self,surf)  
              
        def enter_screen(self):
          app=app_to_install
          self.title.text=app_to_install[3][:50]
          self.val.text=''
          self.overwrite.set(False)
          url=app[2]
          thread=threading.Thread(target=appinfo, 
                 args=(app[1],baseurl+url,self.val,self.title,self.image)).start()
          Screen.enter_screen(self)
        
        def is_ready(self):
           global busy_installing
           return (busy_installing==None and (app_to_install[1].has_key('filename'))) or busy_installing==app_to_install
        def start_install(self):
            global cancel
            global busy_installing
            if busy_installing==None:
              app=app_to_install[3][:50]
              filename=app_to_install[1]['filename']
              url=app_to_install[1]['downloadurl']
              busy_installing=app_to_install
              thread=threading.Thread(target=install_app, 
                 args=(app,self.i.get(),filename,url,self.overwrite.get())).start()
            else:
              cancel=True
              
class BrowseTableScreen(Screen):

	def __init__(self, shell):
		Screen.__init__(self, shell)
		title = Label("Open Handhelds Caanoo repository", font = titlefont)
		recent= Button("Recent", font=buttonfont, action = self.recent,enable=self.is_recent)
                home= Button("Home", font=buttonfont, action = self.home,enable=self.is_home)
                popular= Button("Popular", font=buttonfont, action = self.popular,enable=self.is_pop)
                bbut=Row([recent,home,popular])
                
                table = BrowseTableView(sorted(cur.keys()),height=110,row_height=9,scrolling=True)
		item = Button("Install", font=buttonfont,action = self.button,enable=table.is_item)
		quit= Button("Quit", font=buttonfont,action = self.quit)
                switch = Button("Switch",font=buttonfont, action = table.switch,enable=table.is_dir)
                val=Label(apptext,width=300,font=messfont,align='c')
                table.val=val
                
                content=Column([title, bbut, table, val,Row([item,switch,quit])], spacing = 3,
                                height=220,expand=val)
                
                glab=GlobalLabel('',width=290)
                screen = Column([content,glab], spacing = 3)
		self.add_centered(screen)
                glab.rect=pygame.Rect(5, 223, 290, 11)
                self.shell=shell
                self.table=table
                
        def quit(self):
          ts=threading.enumerate()
          if len(ts)>1:
            alert("this will stop all installs")
          for t in ts[1:]:
            if t.isAlive():
              t._Thread__stop()
          self.shell.quit()              
        
        def recent(self):
          global cur,recent,popular,repo
          self.table.selected_table_row=None
          url=home
          cur=recent
          self.table.rows.scroll_to_item(0)
          thread=threading.Thread(target=load_pop_recent, 
               args=(recent,popular,baseurl+url,self.table.table_data)).start()
          self.table.val.text=''
        def home(self):
          global cur,recent,popular,repo
          self.table.selected_table_row=None
          url=home
          cur=repo
          self.table.rows.scroll_to_item(0)
          thread=threading.Thread(target=switchcategory, 
               args=(cur,baseurl+url,self.table.table_data)).start()
          self.table.val.text=''         
        def popular(self):
          global cur,recent,popular,repo
          self.table.selected_table_row=None
          url=home
          cur=popular
          self.table.rows.scroll_to_item(0)
          thread=threading.Thread(target=load_pop_recent, 
               args=(recent,popular,baseurl+url,self.table.table_data)).start()
          self.table.val.text=''

        def is_home(self):
          return True
        def is_recent(self):        
          return True
        def is_pop(self):
          return True

        def button(self):
          global app_to_install
          if self.table.selected_table_row is not None:
            key=self.table.table_data[self.table.selected_table_row]
            if cur[key][0]=='app':
              app=cur[key]
              app_to_install=app
              self.shell.show_screen(self.shell.install_screen)
	                            

class BrowseTableView(TableView):

	columns = [
		TableColumn("", 280, 'l'),
	]
	header_spacing=0
        header_font=textfont
        font=textfont
        selected_table_row = None
        selected_table_des = ' '*240
        
        def __init__(self, table_data, **kwargs):        
                self.table_data=table_data
                TableView.__init__(self,**kwargs)
                self.rows=self.subwidgets[1]
        
	def num_rows(self):
		return len(self.table_data)
	
	def row_data(self, i):
		return (self.table_data[i][3:],)
	
	def row_is_selected(self, i):
		return self.selected_table_row == i
	
	def click_row(self, i, e):
		self.selected_table_row = i
                try:
                  text=cur[self.table_data[i]][4][:240].ljust(240)
                  self.val.text=text[0:60].strip()+'\n'+text[60:120].strip()+'\n'+text[120:180].strip()+'\n'+text[180:240].strip()
                except:
                  pass
                
        def switch(self):
                global cur
                if self.selected_table_row is not None:
                  key=self.table_data[self.selected_table_row]
                  if cur[key][0]=='cat':
                    self.selected_table_row=None
                    url=cur[key][2]
                    cur=cur[key][1]
                    self.val.text=''
                    self.rows.scroll_to_item(0)
                    thread=threading.Thread(target=switchcategory, 
                              args=(cur,baseurl+url,self.table_data)).start()

        def is_item(self):
          if self.selected_table_row is not None:
            key=self.table_data[self.selected_table_row]
            try:
              if cur[key][0]=='app':
                return True
            except:
              return False   
          return False  

       
        def is_dir(self):
          if self.selected_table_row is not None:
            key=self.table_data[self.selected_table_row]
            try:
              
              if cur[key][0]=='cat':
                return True
            except:
              return False   
          return False  



class BrowseShell(Shell):
	def __init__(self, display):
		Shell.__init__(self, display)
		self.install_screen = AppInstallScreen(self)
                self.menu_screen = BrowseTableScreen(self) 
		self.set_timer(frame_time)
		self.show_menu()
	
	def show_menu(self):
		self.show_screen(self.menu_screen)

def main():
	display = pygame.display.set_mode(screen_size, 0)
	shell = BrowseShell(display)
	shell.redraw_every_frame=True
        shell.run()


repo=dict()
recent=dict()
popular=dict()
filldictfromurl(repo,baseurl+home)
#filldictpopularrecent(popular,recent,baseurl+home)
cur=repo  
app_to_install=None

appinfotext=' \n \n \n \n \n \n'
appscreentitle=' '
apptext=' \n \n \n '
busy_installing=None


main()

